import unittest

from telicent_labels import SecurityLabelBuilder, TelicentSecurityLabelsV2
from telicent_labels.security_labels import Label


class SecurityLabelBuilderTestCase(unittest.TestCase):
    def test_simple_and_group_label_access(self):
        pn = TelicentSecurityLabelsV2.AND_GROUPS.value
        test_group_1 = "urn:telicent:groups:role:doctor"
        test_group_2 = "urn:telicent:groups:role:admin"

        basic_label = pn.create_label(test_group_1)

        self.assertEqual(basic_label, "urn:telicent:groups:role:doctor:and")

        single_group = pn.construct(test_group_1)

        self.assertEqual(single_group, "urn:telicent:groups:role:doctor:and")

        multi_group = pn.construct(test_group_1, test_group_2)

        self.assertEqual(
            multi_group,
            "urn:telicent:groups:role:doctor:and&urn:telicent:groups:role:admin:and",
        )

    def test_simple_or_group_label_access(self):
        pn = TelicentSecurityLabelsV2.OR_GROUPS.value
        test_group_1 = "urn:telicent:groups:role:doctor"
        test_group_2 = "urn:telicent:groups:role:admin"

        basic_label = pn.create_label(test_group_1)

        self.assertEqual(basic_label, "urn:telicent:groups:role:doctor:or")

        single_group = pn.construct(test_group_1)

        self.assertEqual(single_group, "(urn:telicent:groups:role:doctor:or)")

        multi_group = pn.construct(test_group_1, test_group_2)

        self.assertEqual(
            multi_group,
            "(urn:telicent:groups:role:doctor:or|urn:telicent:groups:role:admin:or)",
        )

    def test_simple_single_label_access(self):
        pn = TelicentSecurityLabelsV2.CLASSIFICATION.value
        test_classifcation = "S"

        basic_label = pn.create_label(test_classifcation)

        self.assertEqual(basic_label, "classification=S")

        multi_label = pn.construct(test_classifcation)

        self.assertEqual(multi_label, "classification=S")

    def test_simple_multi_label_access(self):
        pn = TelicentSecurityLabelsV2.PERMITTED_NATIONALITIES.value
        test_nationality_1 = "GBR"
        test_nationality_2 = "NOR"
        basic_label = pn.create_label(test_nationality_1)

        self.assertEqual(basic_label, "permitted_nationalities=GBR")

        multi_label = pn.construct(test_nationality_1, test_nationality_2)

        self.assertEqual(
            multi_label, "(permitted_nationalities=GBR|permitted_nationalities=NOR)"
        )

    def test_security_label_builder(self):
        test_org_1 = "telicent"
        test_org_2 = "nhs"
        test_classification = "S"
        test_classification_1 = "O"
        slb1 = (
            SecurityLabelBuilder()
            .add_multiple(
                TelicentSecurityLabelsV2.PERMITTED_ORGANISATIONS.value,
                test_org_1,
                test_org_2,
            )
            .add(TelicentSecurityLabelsV2.CLASSIFICATION.value, test_classification)
            .build()
        )
        self.assertEqual(
            slb1,
            "((permitted_organisations=telicent|permitted_organisations=nhs)&classification=S)",
        )

        slb2 = (
            SecurityLabelBuilder()
            .add(TelicentSecurityLabelsV2.PERMITTED_ORGANISATIONS.value, test_org_1)
            .add(TelicentSecurityLabelsV2.CLASSIFICATION.value, test_classification_1)
            .build()
        )

        self.assertEqual(
            slb2,
            "((permitted_organisations=telicent)&classification=O)",
        )

        slb3 = (
            SecurityLabelBuilder()
            .add_or_expression(slb1)
            .add_or_expression(slb2)
            .build()
        )

        self.assertEqual(
            slb3,
            "((permitted_organisations=telicent|permitted_organisations=nhs)&classification=S)|"
            "((permitted_organisations=telicent)&classification=O)",
        )

        slb4 = (
            SecurityLabelBuilder()
            .add_multiple(
                TelicentSecurityLabelsV2.PERMITTED_ORGANISATIONS.value,
                test_org_1,
                test_org_2,
            )
            .add(TelicentSecurityLabelsV2.CLASSIFICATION.value, test_classification)
            .add_or_expression(slb2)
            .build()
        )

        self.assertEqual(
            slb4,
            "((permitted_organisations=telicent|permitted_organisations=nhs)&classification=S)|"
            "((permitted_organisations=telicent)&classification=O)",
        )

    def test_security_label_builder_complex(self):
        test_org_1 = "nhs"
        test_classification = "S"
        test_and_group = "urn:telicent:groups:department:vulnerable_care"
        test_or_group_1 = "urn:telicent:groups:role:admin"
        test_or_group_2 = "urn:telicent:groups:role:doctor"
        test_nationality_1 = "GBR"
        test_nationality_2 = "USA"

        slb = (
            SecurityLabelBuilder()
            .add(TelicentSecurityLabelsV2.PERMITTED_ORGANISATIONS.value, test_org_1)
            .add(TelicentSecurityLabelsV2.CLASSIFICATION.value, test_classification)
            .add_multiple(
                TelicentSecurityLabelsV2.OR_GROUPS.value, test_or_group_1, test_or_group_2
            )
            .add(TelicentSecurityLabelsV2.AND_GROUPS.value, test_and_group)
            .add_multiple(
                TelicentSecurityLabelsV2.PERMITTED_NATIONALITIES.value,
                test_nationality_1,
                test_nationality_2,
            )
            .build()
        )
        self.assertEqual(
            slb,
            "((permitted_organisations=nhs)&classification=S&(urn:telicent:groups:role:admin:or|"
            "urn:telicent:groups:role:doctor:or)&urn:telicent:groups:department:vulnerable_care:and&"
            "(permitted_nationalities=GBR|permitted_nationalities=USA))",
        )

    def test_base_label_error(self):
        self.assertRaises(
            NotImplementedError,
            SecurityLabelBuilder().add,
            Label("test", "str"),
            "test",
        )

        self.assertRaises(
            NotImplementedError,
            SecurityLabelBuilder().add_multiple,
            Label("test", "str"),
            "test1",
            "test2",
        )



from datetime import datetime, timezone
from typing import Annotated

from pydantic import AwareDatetime, BaseModel, PlainSerializer

from telicent_labels.security_labels import SecurityLabelBuilder
from telicent_labels.telicentv2 import TelicentSecurityLabelsV2

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


DEFAULT_OPTIONAL_DT = datetime(2023, 12, 14, 0, 0, 0, tzinfo=timezone.utc)
SerialisableDt = Annotated[AwareDatetime, PlainSerializer(lambda x: x.isoformat(), return_type=str, when_used="always")]


class TelicentMixin(BaseModel):

    def build_security_labels(self):
        builder = SecurityLabelBuilder()

        builder.add(TelicentSecurityLabelsV2.CLASSIFICATION.value, self.classification)
        builder.add_multiple(TelicentSecurityLabelsV2.PERMITTED_ORGANISATIONS.value, *self.permittedOrgs)
        builder.add_multiple(TelicentSecurityLabelsV2.PERMITTED_NATIONALITIES.value, *self.permittedNats)
        builder.add_multiple(TelicentSecurityLabelsV2.AND_GROUPS.value, *self.andGroups)
        builder.add_multiple(TelicentSecurityLabelsV2.OR_GROUPS.value, *self.orGroups)

        return builder.build()


class TelicentModel(TelicentMixin):
    apiVersion: str | None = "v1alpha"
    specification: str | None = "UKIC v3.0"
    identifier: str
    classification: str
    permittedOrgs: list[str]
    permittedNats: list[str]
    orGroups: list[str]
    andGroups: list[str]

    createdDateTime: SerialisableDt | None = DEFAULT_OPTIONAL_DT
    originator: str | None = None
    custodian: str | None = None
    policyRef: str | None = None
    dataSet: list[str]
    authRef: list[str]
    dispositionDate: SerialisableDt | None = DEFAULT_OPTIONAL_DT
    dispositionProcess: str | None = None
    dissemination: list[str]

    def build_security_labels(self):
        return super().build_security_labels()



__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class Label:

    """
    A label represents an element within a Security Label Expression.

    It is used to define possible labels within a schema, specifically
    dictating how the label is created and importantly defining a construct
    function to be used as part of the Builder pattern
    """

    def __init__(self, name: str, label_type: str):
        self.name = name
        self.label_type = label_type

    def create_label(self, value):
        return f"{self.name}={value}"

    def construct(self, *values):
        raise NotImplementedError()


class MultiValueLabel(Label):
    """
    The basic implementation for a data label which can take on multiple values
    in an expression, usually using an OR logic
    """

    def construct(self, *values):
        condition = "|".join(map(self.create_label, values))
        return f"({condition})"


class SingleValueLabel(Label):
    """
    The basic implementation for a data label which can take on a single label
    in an expression
    """

    def construct(self, *values):
        if len(values) > 1:
            raise ValueError('SingleValueLabel may only have one value')
        return self.create_label(values[0])


class SecurityLabelHelper:
    def add(self, label: SingleValueLabel, value: str):
        """
        Adds single label expression
        """
        raise NotImplementedError()

    def add_multiple(self, label: MultiValueLabel, *values):
        """
        Adds a single label expression with multiple values, evaluated usually with OR logic
        """
        raise NotImplementedError()

    def add_or_expression(self, expression: str):
        """
        Adds an expression as an or, allowing for complex SecurityLabel expressions
        """
        raise NotImplementedError()

    def build(self) -> str:
        raise NotImplementedError()


class SecurityLabelBuilder(SecurityLabelHelper):
    """
    Core requires all messages to have a security label.

    The Security Label Builder is a helper function to make sure Security Labels created are both valid in terms of
    syntax and valid in terms of standards

    The base label set is the current created Security Label Expression, the or expressions are a way to chain multiple
    expressions together e.g

    """

    def __init__(self):
        self.labels = []
        self.or_expressions = []
        self.used_labels = []

    def add(self, label: Label, value: str):
        if label.name in self.used_labels:
            print(
                "WARNING: Be careful adding multiple expressions with the same label - it could make data inaccessible"
            )
        else:
            self.used_labels.append(label.name)
        self.labels.append(label.construct(value))
        return self

    def add_multiple(self, label: MultiValueLabel, *values):
        if label.name in self.used_labels:
            print(
                "WARNING: Be careful adding multiple expressions with the same label - it could make data inaccessible"
            )
        else:
            self.used_labels.append(label.name)
        self.labels.append(label.construct(*values))
        return self

    def add_or_expression(self, expression: str):
        trimmed = expression.strip()
        if trimmed is None or not len(trimmed) >= 2:
            raise ValueError("Valid security label expression required")
        if trimmed[0] != "(" or trimmed[-1] != ")":
            raise ValueError("Valid security label expression should be enclosed "
                             "with brackets (security-label expression)")
        self.or_expressions.append(trimmed)
        return self

    def build(self):
        open_expression = "("
        expression = "&".join(self.labels)
        close_expression = ")" if len(self.or_expressions) == 0 else ")|"
        base_expression = (
            ""
            if len(expression.strip()) == 0
            else f"{open_expression}{expression}{close_expression}"
        )
        ors = "|".join(self.or_expressions)
        return f"{base_expression}{ors}"



import unittest

from telicent_labels import SecurityLabelBuilder, TelicentLabelsV1


class SecurityLabelBuilderTestCase(unittest.TestCase):
    def test_simple_single_label_access(self):
        pn = TelicentLabelsV1.CLASSIFICATION.value
        test_classifcation = "S"

        basic_label = pn.create_label(test_classifcation)

        self.assertEqual(basic_label, "clearance=S")

        multi_label = pn.construct(test_classifcation)

        self.assertEqual(multi_label, "clearance=S")

    def test_simple_multi_label_access(self):
        pn = TelicentLabelsV1.PERMITTED_NATIONALITIES.value
        test_nationality_1 = "GBR"
        test_nationality_2 = "NOR"
        basic_label = pn.create_label(test_nationality_1)

        self.assertEqual(basic_label, "nationality=GBR")

        multi_label = pn.construct(test_nationality_1, test_nationality_2)

        self.assertEqual(multi_label, "(nationality=GBR|nationality=NOR)")

    def test_security_label_builder(self):
        test_org_1 = "telicent"
        test_org_2 = "nhs"
        test_classification = "S"
        test_classification_1 = "O"
        slb1 = (
            SecurityLabelBuilder()
            .add_multiple(
                TelicentLabelsV1.PERMITTED_ORGANISATIONS.value,
                test_org_1,
                test_org_2,
            )
            .add(TelicentLabelsV1.CLASSIFICATION.value, test_classification)
            .build()
        )
        self.assertEqual(
            slb1,
            "((deployed_organisation=telicent|deployed_organisation=nhs)&clearance=S)",
        )

        slb2 = (
            SecurityLabelBuilder()
            .add(TelicentLabelsV1.PERMITTED_ORGANISATIONS.value, test_org_1)
            .add(TelicentLabelsV1.CLASSIFICATION.value, test_classification_1)
            .build()
        )

        self.assertEqual(
            slb2,
            "((deployed_organisation=telicent)&clearance=O)",
        )

        slb3 = (
            SecurityLabelBuilder()
            .add_or_expression(slb1)
            .add_or_expression(slb2)
            .build()
        )

        self.assertEqual(
            slb3,
            "((deployed_organisation=telicent|deployed_organisation=nhs)&clearance=S)|"
            "((deployed_organisation=telicent)&clearance=O)",
        )

        slb4 = (
            SecurityLabelBuilder()
            .add_multiple(
                TelicentLabelsV1.PERMITTED_ORGANISATIONS.value,
                test_org_1,
                test_org_2,
            )
            .add(TelicentLabelsV1.CLASSIFICATION.value, test_classification)
            .add_or_expression(slb2)
            .build()
        )

        self.assertEqual(
            slb4,
            "((deployed_organisation=telicent|deployed_organisation=nhs)&clearance=S)|"
            "((deployed_organisation=telicent)&clearance=O)",
        )

    def test_security_label_builder_complex(self):
        test_org_1 = "nhs"
        test_classification = "S"

        test_nationality_1 = "GBR"
        test_nationality_2 = "USA"

        slb = (
            SecurityLabelBuilder()
            .add(TelicentLabelsV1.PERMITTED_ORGANISATIONS.value, test_org_1)
            .add(TelicentLabelsV1.CLASSIFICATION.value, test_classification)
            .add_multiple(
                TelicentLabelsV1.PERMITTED_NATIONALITIES.value,
                test_nationality_1,
                test_nationality_2,
            )
            .build()
        )
        self.assertEqual(
            slb,
            "((deployed_organisation=nhs)&clearance=S&(nationality=GBR|nationality=USA))",
        )



from telicent_labels.security_labels import Label, MultiValueLabel, SecurityLabelBuilder, SingleValueLabel
from telicent_labels.telicent_model import TelicentModel
from telicent_labels.telicentv1 import TelicentLabelsV1
from telicent_labels.telicentv2 import TelicentSecurityLabelsV2

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


__all__ = [
    "SecurityLabelBuilder",
    "TelicentSecurityLabelsV2",
    "TelicentLabelsV1",
    "MultiValueLabel",
    "SingleValueLabel",
    "Label",
    "TelicentModel",
]



from enum import Enum

from telicent_labels.security_labels import MultiValueLabel, SingleValueLabel

__license__ = """
Copyright (c) Telicent Ltd.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class AndGroups(MultiValueLabel):

    # Something in here needed to get valid group
    def construct(self, *values):
        # assume value is urn
        condition = "&".join(map(self.create_label, values))
        return condition

    def create_label(self, value: str):
        return value + ":and"


class OrGroups(MultiValueLabel):

    # Something in here needed to get valid group
    def construct(self, *values):
        # assume value is urn
        condition = "|".join(map(self.create_label, values))
        return "(" + condition + ")"

    def create_label(self, value: str):
        return value + ":or"


class TelicentSecurityLabelsV2(Enum):
    """
    Representation of the Telicent model suitable for SecurityLabelbuilder
    """
    PERMITTED_ORGANISATIONS = MultiValueLabel(
        "permitted_organisations", "array"
    )
    PERMITTED_NATIONALITIES = MultiValueLabel(
        "permitted_nationalities", "array"
    )
    CLASSIFICATION = SingleValueLabel("classification", "str")
    AND_GROUPS = AndGroups("and_groups", "groups")
    OR_GROUPS = OrGroups("or_groups", "groups")
